/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.components.treeprocessor;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.component.Recomposable;
import org.apache.avalon.framework.component.ComponentManager;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.avalon.framework.logger.Loggable;

import org.apache.cocoon.components.pipeline.EventPipeline;
import org.apache.cocoon.components.pipeline.StreamPipeline;
import org.apache.cocoon.environment.Environment;
import org.apache.log.Logger;

import java.util.*;

/**
 * The invocation context of <code>ProcessingNode</code>s.
 * <p>
 * This class serves two purposes :
 * <ul><li>Avoid explicit enumeration of all needed parameters in
 *         {@link ProcessingNode#invoke(Environment, InvokeContext)},
 *         thus allowing easier addition of new parameters,
 *     <li>Hold pipelines, and provide "just in time" lookup for them.
 * </ul>
 *
 * @author <a href="mailto:sylvain@apache.org">Sylvain Wallez</a>
 * @version CVS $Id: InvokeContext.java,v 1.1 2002/03/05 08:26:22 sylvain Exp $
 */

public class InvokeContext implements Recomposable, Disposable, Loggable {

    private List mapStack = new ArrayList();

    private StreamPipeline streamPipeline;

    private EventPipeline eventPipeline;

    private boolean isInternalRequest;

    /** The current component manager, as set by the last call to compose() or recompose() */
    private ComponentManager currentManager;

    /** The component manager that was used to get the pipelines */
    private ComponentManager pipelinesManager;

    /** Logger that we will log any messages to */
    private Logger logger;

    /**
     * Create an <code>InvokeContext</code> without existing pipelines. This also means
     * the current request is external.
     */
    public InvokeContext() {
        this.isInternalRequest = false;
    }

    /**
     * Create an <code>InvokeContext</code> with existing pipelines. This also means
     * the current request is internal.
     */
    public InvokeContext(StreamPipeline pipeline, EventPipeline eventPipeline) {
        this.isInternalRequest = true;
        this.streamPipeline = pipeline;
        this.eventPipeline = eventPipeline;
    }

    public void setLogger(Logger logger)
    {
        this.logger = logger;
    }

    public void compose(ComponentManager manager) throws ComponentException {
        this.currentManager = manager;
    }

    public void recompose(ComponentManager manager) throws ComponentException {

        this.currentManager = manager;

        // Recompose pipelines, if any.
        if (this.streamPipeline != null) {
            this.streamPipeline.recompose(manager);
            this.eventPipeline.recompose(manager);
        }
    }

    /**
     * Get the <code>EventPipeline</code>. If none already exists, a set of new
     * pipelines is looked up.
     */
    public final EventPipeline getEventPipeline() throws Exception {
        if (this.eventPipeline == null) {
            setupPipelines();
        }
        return this.eventPipeline;
    }

    /**
     * Get the <code>StreamPipeline</code>. If none already exists, a set of new
     * pipelines is looked up.
     */
    public final StreamPipeline getStreamPipeline() throws Exception {
        if (this.streamPipeline == null) {
            setupPipelines();
        }
        return this.streamPipeline;
    }

    private final void setupPipelines() throws Exception {
        // Keep current manager for proper release
        this.pipelinesManager = this.currentManager;

        // Lookup pipelines
        this.eventPipeline  = (EventPipeline)this.pipelinesManager.lookup(EventPipeline.ROLE);
        this.streamPipeline = (StreamPipeline)this.pipelinesManager.lookup(StreamPipeline.ROLE);
        this.streamPipeline.setEventPipeline(this.eventPipeline);

        // Immediately recompose them to the current manager : pipelines are created
        // by a parent of the current manager, but need to select component in this one,
        // and not its parent.
        this.eventPipeline.recompose(this.pipelinesManager);
        this.streamPipeline.recompose(this.pipelinesManager);
    }

    /**
     * Is this an internal request ?
     */
    public final boolean isInternalRequest() {
        return this.isInternalRequest;
    }

    /**
     * Get the current Map stack used to resolve expressions.
     */
    public final List getMapStack() {
        return this.mapStack;
    }

    /**
     * Push a Map on top of the current Map stack.
     */
    public final void pushMap(Map map) {
        mapStack.add(map);

        if (this.logger.isDebugEnabled()) {
            dumpParameters(getMapStack());
        }
    }

    /**
     * Dumps all sitemap parameters to log
     */
    protected void dumpParameters(List list)
    {
        if (!list.isEmpty()) {
            StringBuffer sb = new StringBuffer();

            sb.append("\nCurrent Sitemap Parameters:\n");
            String path = "";

            for (int i = list.size() - 1; i >= 0; i--) {
                Map map = (Map) list.get(i);
                Iterator keys = map.keySet().iterator();

                while (keys.hasNext()) {
                    String key = (String) keys.next();
                    sb.append("PARAM: '").append(path).append(key);
                    sb.append("' VALUE: '").append(map.get(key)).append("'\n");
                }

                path = "../" + path;
            }

            this.logger.debug(sb.toString());
        }
    }


    /**
     * Pop the topmost element of the current Map stack.
     */
    public final void popMap() {
        mapStack.remove(mapStack.size() - 1);
    }

    /**
     * Release the pipelines, if any, if they were looked up by this context.
     */
    public void dispose() {
        // Release pipelines, if any
        if (!this.isInternalRequest && this.pipelinesManager != null) {

            this.pipelinesManager.release(this.eventPipeline);
            this.pipelinesManager.release(this.streamPipeline);

            this.pipelinesManager = null;
        }
    }
}
