header {
package org.apache.cocoon.flow;
}

options {
    language = Java;
}

class FlowParser extends Parser;

options {
    k = 1;
    exportVocab=jWeave;  // Call its vocabulary "jWeave"
    buildAST = true;     // uses CommonAST by default
    codeGenMakeSwitchThreshold = 2;  // Some optimizations
    codeGenBitsetTestThreshold = 3;
}

tokens {
    DOT; MTHINVOKE; FUNCALL; ARRAYIDX; SEQ; RETURN; ARRAY; DICT; FUNCTION;
    PARAMS; VARARGS;
}

program
    :   exprs EOF!
    ;

exprs!
    :   ( e:expr (SEMI!)+
          {#exprs = #([SEQ], #exprs, #e);}
        )+
    ;

expr
    :   assignment
    |   block           // A sequence of expressions
    |   varDecl         // Variable declaration
    |   function        // Function declaration
    |   "if"^ LPAREN! expr RPAREN! expr
        (
                        // CONFLICT: the old "dangling-else" problem...
                        //           ANTLR generates proper code matching
                        //           as soon as possible.  Hush warning.
            options {
                warnWhenFollowAmbig = false;
            }:
            "else" expr
        )?

        // The following expressions should be rewritten using macros,
        // when they become available
    |!  "return"^ return:expr
        {#expr = #([RETURN], return); }
    |   "try"^ expr
        (
            // CONFLICT: Similar with the dangling else
            options {
                warnWhenFollowAmbig = false;
            }:
            handler
        )+
    |   "throw"^ expr
    |   "import"^ identifierStar
    |   "while"^ LPAREN! expr RPAREN! expr
    |   "switch"^ LPAREN! expr RPAREN! casesGroup
    |   "for"^ LPAREN! exprList SEMI! exprList SEMI! exprList RPAREN! expr
    |   "foreach"^ LPAREN! IDENT "in" expr RPAREN! expr
    |   "do"^ expr "while" LPAREN! expr RPAREN!
    ;

exprList
    :   assignment
        (
            // CONFLICT: Similar with the dangling else
            options {
                warnWhenFollowAmbig = false;
            }:
            COMMA! assignment
        )*
    ;

// A function is either named or unnamed. A named function is //
// equivalent to a variable declaration whose value is an unnamed
// function.
function!
    :   "function"^ (n:IDENT)? p:parameters b:block
        {#function = #([FUNCTION, "function"], n, #([PARAMS], p), b);}
    ;

parameters!
    :   LPAREN!
          (funParam (COMMA! funParam)*)?
          (v:"*args" {#parameters.addChild(#[VARARGS, "*args"]);})? 
        RPAREN!
    ;

funParam
    :   IDENT (ASSIGN^ (IDENT|constant))?
    ;

// A sequence of expressions. The value of the block is the value of
// the last expression evaluated inside the block.
block
    :   LCURLY! exprs RCURLY!
    ;

// Variable declaration
varDecl
    :   "var"^ exprList
    ;

casesGroup
    :   (   // CONFLICT: to which case group do the exprs bind?
            //           ANTLR generates proper code: it groups the
            //           many "case"/"default" labels together then
            //           follows them with the exprs
            options {
                warnWhenFollowAmbig = false;
            }
        :
            aCase
        )+
        caseSList
    ;

aCase
    :   ("case" expr | "default") COLON!
    ;

caseSList
    :   exprs
    ;

// Handler for a try expression
handler
    :   "catch" LPAREN! expr IDENT RPAREN! expr
    |   "finally" expr
    ;

identifierStar
    :   IDENT
        (DOT! (IDENT|STAR))*
    ;

assignment
    :   logicalOr
        (   ( ASSIGN^
            | PLUS_ASSIGN^
            | MINUS_ASSIGN^
            | STAR_ASSIGN^
            | DIV_ASSIGN^
            | MOD_ASSIGN^
            | BAND_ASSIGN^
            | BXOR_ASSIGN^
            | BOR_ASSIGN^
            )
            assignment
        )?
    ;

logicalOr
    :   logicalAnd (LOR^ logicalAnd)*
    ;

logicalAnd
    :   inclusiveOr (LAND^ inclusiveOr)*
    ;

inclusiveOr
    :   exclusiveOr (BOR^ exclusiveOr)*
    ;

exclusiveOr
    :   and (BXOR^ and)*
    ;

and
    :   equality (BAND^ equality)*
    ;

equality
    :   relational ((NOT_EQUAL^ | EQUAL^) relational)*
    ;

relational
    :   additive
        (   (   ( LT^
                | GT^
                | LE^
                | GE^
                )
                additive
            )*
        )
    ;

additive
    :   multiplicative ((PLUS^ | MINUS^) multiplicative)*
    ;

multiplicative
    :   unary ((STAR^ | DIV^ | MOD^) unary)*
    ;

unary
    :   INC^ unary
    |   DEC^ unary
    |   MINUS^ unary
    |   PLUS^ unary
    |   unaryNotPlusMinus
    |   arrayOrDict
    ;

unaryNotPlusMinus
    :   BNOT^ unary
    |   LNOT^ unary
    |   postfix
    ;

postfix
    :   primary
        (
            // Array indexing
            LBRACK! a:assignment RBRACK!
            {#postfix = #([ARRAYIDX], a);}

            // Method invocation
        |   DOT! LPAREN! ma:exprList RPAREN!
            {#postfix = #([MTHINVOKE], ma);}

            // Function invocation
        |   LPAREN! fa:exprList RPAREN!
            {#postfix = #([FUNCALL], fa);}
        )*
    ;

primary
    :   IDENT
    |   constant
    |   LPAREN! assignment RPAREN!
    ;

constant
    :   NUMBER
    |   CHAR_LITERAL
    |   STRING_LITERAL
    |   "true"
    |   "false"
    |   "null"
    ;

// An array or a dictionary:
//   [a, b, c] -> array
//   [a: 1, b: 2, c: 3] -> dictionary
// Use a syntactic predicate to disambiguate on ':'
arrayOrDict!
    :   (LBRACK! expr COLON!)=>
        LBRACK! key:expr COLON! value:expr
        (COMMA! keys:expr COLON! values:expr)* RBRACK!
        {#arrayOrDict = #([DICT, "dict"], key, value, ([DICT], keys, values));}
    |   LBRACK! elems:exprList RBRACK!
        {#arrayOrDict = #([ARRAY, "array"], elems);}
    ;



class FlowLexer extends Lexer;

options {
    k = 2;
    exportVocab=jWeave;    // Call its vocabulary "jWeave"
    testLiterals=false;    // don't automatically test for literals
}

QUESTION                :       '?'     ;
LPAREN                  :       '('     ;
RPAREN                  :       ')'     ;
LBRACK                  :       '['     ;
RBRACK                  :       ']'     ;
LCURLY                  :       '{'     ;
RCURLY                  :       '}'     ;
COLON                   :       ':'     ;
COMMA                   :       ','     ;
ASSIGN                  :       '='     ;
EQUAL                   :       "=="    ;
LNOT                    :       '!'     ;
BNOT                    :       '~'     ;
NOT_EQUAL               :       "!="    ;
DIV                     :       '/'     ;
DIV_ASSIGN              :       "/="    ;
PLUS                    :       '+'     ;
PLUS_ASSIGN             :       "+="    ;
INC                     :       "++"    ;
MINUS                   :       '-'     ;
MINUS_ASSIGN            :       "-="    ;
DEC                     :       "--"    ;
STAR                    :       '*'     ;
STAR_ASSIGN             :       "*="    ;
MOD                     :       '%'     ;
MOD_ASSIGN              :       "%="    ;
GE                      :       ">="    ;
GT                      :       ">"     ;
LE                      :       "<="    ;
LT                      :       '<'     ;
BXOR                    :       '^'     ;
BXOR_ASSIGN             :       "^="    ;
BOR                     :       '|'     ;
BOR_ASSIGN              :       "|="    ;
LOR                     :       "||"    ;
BAND                    :       '&'     ;
BAND_ASSIGN             :       "&="    ;
LAND                    :       "&&"    ;
SEMI                    :       ';'     ;

protected
PANY    :   ('a'..'z'|'A'..'Z'|'_');

protected
SPECIAL :   ('@'|'#'|'$');

protected
ANY     :   (PANY|COLON|'0'..'9');

IDENT
	options {testLiterals=true;}
        :   (PANY (ANY)*) | (SPECIAL (ANY)+);

//
// The following are stolen from java.g in the ANTLR distribution.
//
WS  :   (   ' '
        |   '\t'
        |   '\f'
            // handle newlines
        |   (   options {generateAmbigWarnings=false;}
            :   "\r\n"  // Evil DOS
            |   '\r'    // Macintosh
            |   '\n'    // Unix (the right way)
            )
            { newline(); }
        )+
        { _ttype = Token.SKIP;}
    ;

// Single-line comments
SL_COMMENT
    :   "//"
        (~('\n'|'\r'))* ('\n'|'\r'('\n')?)
        {$setType(Token.SKIP); newline();}
    ;

// multiple-line comments
ML_COMMENT
    :   "/*"
        (   /*  '\r' '\n' can be matched in one alternative or by matching
                '\r' in one iteration and '\n' in another.  I am trying to
                handle any flavor of newline that comes in, but the language
                that allows both "\r\n" and "\r" and "\n" to all be valid
                newline is ambiguous.  Consequently, the resulting grammar
                must be ambiguous.  I'm shutting this warning off.
             */
            options {
                generateAmbigWarnings=false;
            }
        :
            { LA(2)!='/' }? '*'
        |   '\r' '\n'       {newline();}
        |   '\r'            {newline();}
        |   '\n'            {newline();}
        |   ~('*'|'\n'|'\r')
        )*
        "*/"
        {$setType(Token.SKIP);}
    ;


// character literals
CHAR_LITERAL
    :   '\'' ( ESC | ~'\'' ) '\''
    ;

// string literals
STRING_LITERAL
    :   '"' (ESC|~('"'|'\\'))* '"'
    ;


// escape sequence -- note that this is protected; it can only be called
//   from another lexer rule -- it will not ever directly return a token to
//   the parser
// There are various ambiguities hushed in this rule.  The optional
// '0'...'9' digit matches should be matched here rather than letting
// them go back to STRING_LITERAL to be matched.  ANTLR does the
// right thing by matching immediately; hence, it's ok to shut off
// the FOLLOW ambig warnings.
protected
ESC
    :   '\\'
        (   'n'
        |   'r'
        |   't'
        |   'b'
        |   'f'
        |   '"'
        |   '\''
        |   '\\'
        |   ('u')+ HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT 
        |   ('0'..'3')
            (
                options {
                    warnWhenFollowAmbig = false;
                }
            :   ('0'..'7')
                (   
                    options {
                        warnWhenFollowAmbig = false;
                    }
                :   '0'..'7'
                )?
            )?
        |   ('4'..'7')
            (
                options {
                    warnWhenFollowAmbig = false;
                }
            :   ('0'..'9')
            )?
        )
    ;


// hexadecimal digit (again, note it's protected!)
protected
HEX_DIGIT
    :   ('0'..'9'|'A'..'F'|'a'..'f')
    ;


// a dummy rule to force vocabulary to be all characters (except special
//   ones that ANTLR uses internally (0 to 2)
protected
VOCAB
    :   '\3'..'\377'
    ;


// a numeric literal
NUMBER
    {boolean isDecimal=false;}
    :   '.' {_ttype = DOT;}
            (('0'..'9')+ (EXPONENT)? (FLOAT_SUFFIX)? { _ttype = NUMBER; })?
    |   (   '0' {isDecimal = true;} // special case for just '0'
            (   ('x'|'X')
                (                                           // hex
                    // the 'e'|'E' and float suffix stuff look
                    // like hex digits, hence the (...)+ doesn't
                    // know when to stop: ambig.  ANTLR resolves
                    // it correctly by matching immediately.  It
                    // is therefor ok to hush warning.
                    options {
                        warnWhenFollowAmbig=false;
                    }
                :   HEX_DIGIT
                )+
            |   ('0'..'7')+                                 // octal
            )?
        |   ('1'..'9') ('0'..'9')*  {isDecimal=true;}       // non-zero decimal
        )
        (   ('l'|'L')
        
        // only check to see if it's a float if looks like decimal so far
        |   {isDecimal}?
            (   '.' ('0'..'9')* (EXPONENT)? (FLOAT_SUFFIX)?
            |   EXPONENT (FLOAT_SUFFIX)?
            |   FLOAT_SUFFIX
            )
            { _ttype = NUMBER; }
        )?
    ;


// a couple protected methods to assist in matching floating point numbers
protected
EXPONENT
    :   ('e'|'E') ('+'|'-')? ('0'..'9')+
    ;


protected
FLOAT_SUFFIX
    :   'f'|'F'|'d'|'D'
    ;


// Local Variables:
// tab-width: 8
// indent-tabs-mode: nil
// End:
