/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.scheme.sitemap;

import java.lang.Math;
import java.security.SecureRandom;
import java.util.HashMap;
import sisc.ContinuationException;
import sisc.Interpreter;
import sisc.ModuleAdapter;
import sisc.Util;
import sisc.data.ImmutableString;
import sisc.data.Value;

/**
 * Manages the continuation objects.
 *
 * @author <a href="mailto:ovidiu@cup.hp.com">Ovidiu Predescu</a>
 * @since February 8, 2002
 */
public class ContinuationsManager extends CocoonModule
{
  static protected int CONTINUATION_ID_LENGTH = 20;

  public String getModuleName()
  {
    return "Continuations Manager";
  }

  public float getModuleVersion()
  {
    return 1.0f;
  }

  public static final int
    REGISTER = 1,
    GET = 2,
    REMOVE = 3;

  /**
   * Random generator for continuation identifiers.
   */
  protected SecureRandom random = null;

  /**
   * Array of bytes used internally to generate the continuation
   * identifier.
   */
  protected byte[] bytes;

  /**
   * Association between string identifiers and continuation objects.
   */
  protected HashMap continuations = new HashMap();

  public ContinuationsManager()
    throws Exception
  {
    random = SecureRandom.getInstance("SHA1PRNG");
    random.setSeed(System.currentTimeMillis());
    bytes = new byte[CONTINUATION_ID_LENGTH];

    define("continuations:register", REGISTER);
    define("continuations:get", GET);
    define("continuations:remove", REMOVE);
  }

  /**
   * The SISC evaluator function for this module. Executes the actual
   * Java code corresponding to the Scheme functions defined by the
   * module.
   *
   * @param primid an <code>int</code> value
   * @param r an <code>Interpreter</code> value
   * @return a <code>Value</code> value
   * @exception ContinuationException if an error occurs
   */
  public Value eval(int primid, Interpreter r)
    throws ContinuationException
  {
    switch (r.vlr.length) {
      // One argument functions
    case 1:
      switch (primid) {
      case REGISTER:
        return registerContinuation(r.vlr[0]);
      case GET:
        return getContinuationFromId(r.vlr[0]);
      case REMOVE:
        return removeContinuationId(r.vlr[0]);
      }
    }

    throw new RuntimeException("Invalid number of arguments to function "
                               + r.acc + ", got " + r.vlr.length);
  }

  /**
   * Registers a continuation object and returns an identifier
   * associated with it.
   *
   * @param kont a continuation object
   * @return a <code>Value</code> value, representing the string identifier
   */
  public Value registerContinuation(Value kont)
  {
    Value kontId;

    char[] result = new char[bytes.length * 2];
    String continuationId = null;
    
    synchronized(continuations) {
      do {
        random.nextBytes(bytes);
    
        for (int i = 0; i < CONTINUATION_ID_LENGTH; i++) {
          byte ch = bytes[i];      
          result[2 * i] = Character.forDigit(Math.abs(ch >> 4), 16);
          result[2 * i + 1] = Character.forDigit(Math.abs(ch & 0x0f), 16);
        }
        continuationId = new String(result);
      } while (continuations.containsKey(continuationId));
      continuations.put(continuationId, kont);
    }

    return new ImmutableString(continuationId);
  }

  /**
   * Given a Scheme string identifier, return the corresponding
   * continuation.
   *
   * @param id the string identifier
   * @return the continuation object, or <code>#f</code> if no such
   * continuation is registered.
   */
  public Value getContinuationFromId(Value id)
  {
    String continuationId = string(id);
    Value kont = (Value)continuations.get(continuationId);
    return (kont != null ? kont : Util.FALSE);
  }

  /**
   * Given a continuation id, remove the associated continuation from
   * the store.
   *
   * @param id the string identifier for the continuation
   * @return the continuation object removed
   */
  public Value removeContinuationId(Value id)
  {
    synchronized(continuations) {
      return (Value)continuations.remove(string(id));
    }
  }
}
