/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================

 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.

 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:

 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.

 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.

 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache
 Software Foundation, please see <http://www.apache.org/>.

*/
package org.apache.cocoon.sunshine.connector;

import java.io.IOException;

import org.w3c.dom.DocumentFragment;

import org.apache.avalon.excalibur.source.SourceParameters;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;

import org.xml.sax.ContentHandler;
import org.xml.sax.ext.LexicalHandler;

/**
 *  The Component for loading and saving xml to external resources.
 *
 *  An external resource is specified by a type and an identifier. The type
 *  gives the connection information (file, http etc) and the identifier
 *  identifies the resource itself (filename, http-address etc).
 *
 * @author <a href="mailto:cziegeler@s-und-n.de">Carsten Ziegeler</a>
 * @version CVS $Id: ResourceConnector.java,v 1.2 2002/02/22 06:57:17 cziegeler Exp $
*/
public interface ResourceConnector
extends Component {

    /** Role name */
    static public final String ROLE = "org.apache.cocoon.sunshine.connector.ResourceConnector";

    /** No type information specified. */
    static public final int RESOURCE_TYPE_NONE  = 0;

    /**
     *  Resource is an URI.
     *  The resource is specified by an uri, local or remote. If it is a
     *  remote resource the http protocol is used. For local resources the
     *  sunShine pipelines are invoked.
     *  The <code>typeParameter</code> "method" specifies the http method (POST,
     *  GET etc) - standard is POST if parameters are used, without parameters
     *  it is GET.
     *  The <code>typeParameter</code> "followRedirects" specifies if the
     *  connection should follow redirects.
     *  URLRewriting can be turned on by specifying the parameters defined
     *  for the <code>URLRewriter</code>.
     *  If an URI is saved (e.g. saveXML) this is the same as loading from
     *  this URI and specifying a resource parameter named "content" with
     *  the document fragment as a string.
     */
    static public final int RESOURCE_TYPE_URI   = 1;
    /**
     * Resource is a class.
     * The class must conform to the XMLLoader or the XMLSaver interface.
     */
    static public final int RESOURCE_TYPE_CLASS = 2;
    /**
     * Resource is a file.
     */
    static public final int RESOURCE_TYPE_FILE  = 3;

    /**
     * Load an XML Fragment from a resource.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     *
     * @return The loaded data.
     */
    public DocumentFragment loadXML(int                resourceType,
                                    Parameters         typeParameters,
                                    String             resourceIdentifier,
                                    SourceParameters   resourceParameters)
    throws ProcessingException;

    /**
     * Save an XML Fragment to a resource.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     * @param fragment            The data to be saved.
     */
    public void saveXML(int                resourceType,
                        Parameters         typeParameters,
                        String             resourceIdentifier,
                        SourceParameters   resourceParameters,
                        DocumentFragment   fragment)
    throws ProcessingException;

    /**
     * Stream an XML resource.
     * This is like loading the resource and then streaming it, but without
     * the Resource->SAX->DOM->SAX transformation, its simply: Resource->SAX
     * All SAX events are send except of the startDocument() and
     * endDocument() event.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     * @param contentHandler      The SAX ContentHandler receiving the events.
     * @param lexicalHandler      The SAX LexicalHandler receiving the events.
     */
    public void streamXML(int                resourceType,
                          Parameters         typeParameters,
                          String             resourceIdentifier,
                          SourceParameters   resourceParameters,
                          ContentHandler     contentHandler,
                          LexicalHandler     lexicalHandler)
    throws ProcessingException;

    /**
     * Get the content of a resource.
     * This method can be used to get any data from a resource, not only xml.
     * The data is represented by a String.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     */
    public String getContent(int                resourceType,
                             Parameters         typeParameters,
                             String             resourceIdentifier,
                             SourceParameters   resourceParameters)
    throws ProcessingException;

    /**
     * Save a String to a resource.
     *
     * @param resourceType        The type of the connection.
     * @param typeParameters      Connection Parameters.
     * @param resourceIdentifier  The resource name (filename, http address etc)
     * @param resourceParameters  Parameters for the resource.
     * @param content             The data to be saved.
     */
    public void saveContent(int                resourceType,
                            Parameters         typeParameters,
                            String             resourceIdentifier,
                            SourceParameters   resourceParameters,
                            String             content)
    throws ProcessingException;
}
