/*

 ============================================================================
                   The Apache Software License, Version 1.1
 ============================================================================
 
 Copyright (C) 1999-2002 The Apache Software Foundation. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:
 
 1. Redistributions of  source code must  retain the above copyright  notice,
    this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.
 
 3. The end-user documentation included with the redistribution, if any, must
    include  the following  acknowledgment:  "This product includes  software
    developed  by the  Apache Software Foundation  (http://www.apache.org/)."
    Alternately, this  acknowledgment may  appear in the software itself,  if
    and wherever such third-party acknowledgments normally appear.
 
 4. The names "Apache Cocoon" and  "Apache Software Foundation" must  not  be
    used to  endorse or promote  products derived from  this software without
    prior written permission. For written permission, please contact
    apache@apache.org.
 
 5. Products  derived from this software may not  be called "Apache", nor may
    "Apache" appear  in their name,  without prior written permission  of the
    Apache Software Foundation.
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 APACHE SOFTWARE  FOUNDATION  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT,
 INDIRECT, INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLU-
 DING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Apache Software  Foundation and was  originally created by
 Stefano Mazzocchi  <stefano@apache.org>. For more  information on the Apache 
 Software Foundation, please see <http://www.apache.org/>.
 
*/

package org.apache.cocoon.transformation.pagination;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import org.apache.cocoon.Modifiable;
import org.apache.cocoon.util.ResizableContainer;

/**
 * Interprets the pagesheet rules to perform pagination.
 * 
 * FIXME (SM): this code sucks! It was done to show the concept of 
 *             rule driven pagination (which I find very nice) but
 *             it needs major refactoring in order to be sufficiently
 *             stable to allow any input to enter without breaking
 *             SAX well-formness. I currently don't have the time to make
 *             it any better (along with implementing the char-based rule
 *             that is mostly useful for text documents) but if you want
 *             to blast the code and rewrite it better, you'll make me happy :)
 *
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @version CVS $Revision: 1.4 $ $Date: 2002/02/22 16:48:56 $
 */
 
/* 

This is an example pagesheet to show the power of this:

  <?xml version="1.0"?>
  <pagesheet xmlns="http://apache.org/cocoon/paginate/1.0">
   <items>
    <group name="pictures" element="file" namespace="http://apache.org/cocoon/directory/2.0"/>
   </items>
   <rules page="1">
    <count type="element" name="file" namespace="http://apache.org/cocoon/directory/2.0" num="16" store="yes"/>
     <link type="unit" num="2"/>
     <link type="range" value="10"/>
   </rules>
   <rules>
    <count type="element" name="file" namespace="http://apache.org/cocoon/directory/2.0" num="16" store="yes"/>
     <link type="unit" num="5"/>
     <link type="range" value="20"/>
   </rules>
  </pagesheet>

which indicates that:

 1) there is one item group called "picture" and each item is given by the 
    element "file" of the namespace "http://apache.org/cocoon/directory/2.0".
    
 2) for the first page, the pagination rules indicate that there are two unit
    links (two above and two below, so linking to page -2 -1 0 +1 +2) and 
    range links have value 10 (so they link to page -10 and +10).
 
 3) for the rest of the pages, there are three unit links (-3 -2 -1 0 +1 +2 +3)
    and range goes 20 (so +20 and -20).
    
*/

public class Pagesheet extends DefaultHandler implements Cloneable, Modifiable {

    private int level = 0;
    private int pg = 0;
    private long lastModified;
    private Pagesheet pagesheet;
    private PageRules rules;

    ResizableContainer pageRules;
    ResizableContainer pages;

    Map itemGroupsPerName;
    Map itemGroupsPerElement;
    Map itemListsPerName;
    Map itemListsPerElement;

    private int currentPage = 1;
    private int currentPageElementStart = 1;
    private int currentPageElementEnd = 1;
    private int counter = 0;
    private int descendant = 0;

    class Page {
        public int elementStart;
        public int elementEnd;

        public Page(int start, int end) {
            this.elementStart = start;
            this.elementEnd = end;
        }

        public boolean validInPage(int counter) {
            return (this.elementStart <= counter) && (counter <= this.elementEnd);
        }
    }

    class ItemList extends ArrayList {

        public ItemList(int capacity) {
            super(capacity);
        }

        public void addItem(int page) {
            this.add(new Integer(page));
        }

        public int getPageForItem(int item) {
            Integer i = (Integer) this.get(item-1);
            return (i == null) ? 0 : i.intValue();
        }

        public boolean valid(int item) {
            return (item == this.size());
        }
    }

    public Pagesheet() {
        this.pages = new ResizableContainer(2);
    }

    private Pagesheet(ResizableContainer rules, Map itemGroupsPerName, Map itemGroupsPerElement) {
        this.pageRules = rules;
        this.itemGroupsPerName = itemGroupsPerName;
        this.itemGroupsPerElement = itemGroupsPerElement;

        this.pages = new ResizableContainer(5);
        
        if ((this.itemGroupsPerName != null) && (this.itemGroupsPerElement != null)) {
            this.itemListsPerName = new HashMap(itemGroupsPerName.size());
            this.itemListsPerElement = new HashMap(itemGroupsPerName.size());
    
            Iterator iter = itemGroupsPerName.values().iterator();
            for (int i = 0; iter.hasNext(); i++) {
                ItemGroup group = (ItemGroup) iter.next();
                ItemList list = new ItemList(10);
                this.itemListsPerName.put(group.getName(),list);
                this.itemListsPerElement.put(group.getElementURI()+group.getElementName(),list);
            }
        }
    }

    // --------------- interprets the pagesheet document ----------------

    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        if (!uri.equals(Paginator.PAGINATE_URI)) {
            throw new SAXException("The pagesheet's namespace is not supported.");
        }
    }

    public void startElement(String uri, String loc, String raw, Attributes a) throws SAXException {
        level++;
        switch (level) {
            case 1:
                if (loc.equals("pagesheet")) {
                    pagesheet = new Pagesheet();
                    return;
                }
            case 2:
                if (loc.equals("rules")) {
                    if (this.pageRules == null) this.pageRules = new ResizableContainer(2);
                    String key = a.getValue("page");
                    if (key != null) {
                        try {
                            pg = Integer.parseInt(key);
                        } catch (NumberFormatException e) {
                            throw new SAXException("Syntax error: the attribute 'rules/@page' must contain a number");
                        }
                    } else {
                        pg = 0;
                    }
                    rules = new PageRules();
                    return;
                } else if (loc.equals("items")) {
                    if (this.itemGroupsPerName == null) this.itemGroupsPerName = new HashMap(2);
                    if (this.itemGroupsPerElement == null) this.itemGroupsPerElement = new HashMap(2);
                }
            case 3:
                if (loc.equals("count")) {
                    if (a.getValue("type").equals("element")) {
                        rules.elementName = a.getValue("name");
                        rules.elementURI = a.getValue("namespace");
                        rules.store = (a.getValue("store") != null);
                        try {
                            rules.elementCount = Integer.parseInt(a.getValue("num"));
                        } catch (NumberFormatException e) {
                            throw new SAXException("Syntax error: the attribute 'count/@num' must contain a number");
                        }
                    } else if (a.getValue("type").equals("chars")) {
                        try {
                            rules.charCount = Integer.parseInt(a.getValue("num"));
                        } catch (NumberFormatException e) {
                            throw new SAXException("Syntax error: the attribute 'count/@num' must contain a number.");
                        }
                    } else {
                        throw new SAXException("Syntax error: count type not supported.");
                    }
                } else if (loc.equals("link")) {
                    if (a.getValue("type").equals("unit")) {
                        try {
                            rules.unitLinks = Integer.parseInt(a.getValue("num"));
                        } catch (NumberFormatException e) {
                            throw new SAXException("Syntax error: the attribute 'link/@num' must contain a number.");
                        }
                    } else if (a.getValue("type").equals("range")) {
                        try {
                            rules.rangeLink = Integer.parseInt(a.getValue("value"));
                        } catch (NumberFormatException e) {
                            throw new SAXException("Syntax error: the attribute 'link/@value' must contain a number.");
                        }
                    } else {
                        throw new SAXException("Syntax error: link type not supported.");
                    }
                } else if (loc.equals("group")) {
                    String name = a.getValue("name");
                    String elementName = a.getValue("element");
                    String elementURI = a.getValue("namespace");
                    if (name == null) throw new SAXException("Syntax error: the attribute 'group/@name' must be present.");
                    if (elementName == null) throw new SAXException("Syntax error: the attribute 'group/@element' must be present.");
                    ItemGroup group = new ItemGroup(name, elementURI, elementName);
                    this.itemGroupsPerName.put(name, group);
                    this.itemGroupsPerElement.put(elementURI+elementName, group);
                }
                return;
        }
        throw new SAXException("Syntax error: element " + raw + " is not recognized or is misplaced.");
    }

    public void endElement(String uri, String loc, String raw) throws SAXException {
        level--;
        if (loc.equals("rules")) {
            pageRules.set(pg, rules);
        }
    }

    public void endDocument() throws SAXException {
        if (pageRules.size() == 0) throw new SAXException("Pagesheet must contain at least a set of pagination rules.");
        if (pageRules.get(0) == null) throw new SAXException("Pagesheet must contain the global pagination rules.");
    }

    // --------------- process the received element events ----------------

    public void processStartElement(String uri, String name) {
        PageRules rules = getPageRules(currentPage);

        if (rules.match(name,uri)) {
            counter++;
            descendant++;
            if (counter == currentPageElementStart) {
                currentPageElementEnd = currentPageElementStart + rules.elementCount - 1;
                pages.set(currentPage, new Page(currentPageElementStart, currentPageElementEnd));
            }
            if (counter == currentPageElementEnd) {
                currentPage++;
                currentPageElementStart = currentPageElementEnd + 1;
            }
        }

        if (itemGroupsPerElement != null) {
            String qname = uri + name;
            ItemGroup group = (ItemGroup) this.itemGroupsPerElement.get(qname);
            if ((group != null) && (group.match(uri))) {
                ItemList list = (ItemList) this.itemListsPerElement.get(qname);
                if (list != null) list.addItem(currentPage);
            }
        }
    }

    public void processEndElement(String uri, String name) {
        PageRules rules = getPageRules(currentPage);
        if (rules.match(name,uri)) {
            descendant--;
        }
    }

    public void processCharacters(char[] ch, int index, int len) {
        PageRules rules = getPageRules(currentPage);
        // FIXME (SM): do nothing for now but this is the place where
        // we should process the character to do char-based pagination
        // instead of element-based pagination which is much more
        // useful for text-based documents.
    }

    // --------------- return the pagination information ----------------

    public boolean isInPage(int page, int item, String itemGroup) {
        return ((descendant == 0) || valid(page, item, itemGroup));
    }

    public int getTotalPages() {
        return currentPage;
    }

    public int getTotalItems(String itemGroup) {
        ItemList list = (ItemList) this.itemListsPerName.get(itemGroup);
        return (list == null) ? 0 : list.size();
    }

    public int getPageForItem(int item, String itemGroup) {
        ItemList list = (ItemList) this.itemListsPerName.get(itemGroup);
        return (list == null) ? 0 : list.getPageForItem(item);
    }

    public boolean store(String elementURI, String elementName) {
        PageRules rules = getPageRules(currentPage);
        return rules.store;        
    }

    public int itemCount(String elementURI, String elementName) {
        ItemList list = (ItemList) this.itemListsPerElement.get(elementURI+elementName);
        return (list == null) ? 0 : list.size();
    }

    public String getItemGroupName(String elementURI, String elementName) {
        return ((ItemGroup) this.itemGroupsPerElement.get(elementURI+elementName)).getName();
    }

    // ---------------- miscellaneous methods ----------------------------

    private boolean valid(int page, int item, String itemGroup) {
        if (item == 0) {
            Page p = (Page) pages.get(page);
            return (p != null) && (p.validInPage(counter));
        } else {
            ItemList list = (ItemList) this.itemListsPerName.get(itemGroup);
            return (list != null) && (list.valid(item));
        }
    }

    public PageRules getPageRules(int page) {
        PageRules p = (PageRules) pageRules.get(page);
        return (p != null) ? p : (PageRules) pageRules.get(0);
    }

    public void setLastModified(long lastModified) {
        this.lastModified = lastModified;
    }

    public boolean modifiedSince(long date) {
        return (date != this.lastModified);
    }

    public Object clone() {
        return new Pagesheet(pageRules, itemGroupsPerName, itemGroupsPerElement);
    }
}
